# This script contains the functions for calculating PRTP in the paper
# This includes LATE and LATEstar (LATEst)
# Estimation done with six groups

library("lpSolve")

###############
# Aim: to get group proportions given defiers and switching groups
empp <- function(df,t=t,z=z){
  t0z0 <- as.numeric(df$t==0&df$z==0)
  t0z1 <- as.numeric(df$t==0&df$z==1)
  t1z0 <- as.numeric(df$t==1&df$z==0)
  t1z1 <- as.numeric(df$t==1&df$z==1)
  empz1 <- mean(df$z)
  
  p00 <- mean(t0z0)/(1-empz1)
  p10 <- mean(t1z0)/(1-empz1)
  p01 <- mean(t0z1)/empz1
  p11 <- mean(t1z1)/empz1
  
  c(p00,p10,p01,p11)
}

## Counterfactual: increase treatment propensity
# Specify qD, Pr(CA|C), and Pr(NC|N) to get proportions of 6 groups
# If we want LATE, set NCN=0
PrG6fn <- function(df,qD=0,CAC=0,NCN=0,t=t,z=z){
  pobs <- empp(df)
  p00 <- pobs[1]; p10 <- pobs[2]; p01 <- pobs[3]; p11 <- pobs[4]
  
  natPrD <- min(p10,p01)
  if(qD >= natPrD) qD <- natPrD
  qA <- 1 - p00 - qD
  qN <- 1 - p11 - qD
  qC <- p00 + p11 - 1 + qD
  
  c(qA,CAC*qC,(1-CAC)*qC,qD,NCN*qN,(1-NCN)*qN)
}


##############
# Linear Program 
LATEstBounds <- function(df,PrG6,t=t,z=z,y=y,K=c(0,1),
                         NegTE=TRUE, minTE=-.3) {
  pobs <- empp(df)
  p00 <- pobs[1]; p10 <- pobs[2]; p01 <- pobs[3]; p11 <- pobs[4]
  dt0z0 <- subset(df,t==0&z==0)
  dt0z1 <- subset(df,t==0&z==1)
  dt1z0 <- subset(df,t==1&z==0)
  dt1z1 <- subset(df,t==1&z==1)
  pz <- mean(df$z)
  
  y00 <- mean(dt0z0$y)
  y01 <- mean(dt0z1$y)
  y10 <- mean(dt1z0$y)
  y11 <- mean(dt1z1$y)
  
  qA <- PrG6[1]; qCA <- PrG6[2]; qCC <- PrG6[3]; qD <- PrG6[4]; 
  qNC <- PrG6[5]; qNCp <- PrG6[6]; 
  
  # Force max and min K
  if (K[2]==Inf) K <- c(min(df$y),max(df$y))
  
  ## Build LP
  # Choose Objective
  # Treatment effect on new compliers
  f.obj <- c(rep(0,4),qCC,-qCC,rep(0,2),qNC,-qNC,rep(0,2))/(qCC+qNC)
  
  
  
  # Set matrix corresponding to coefficients of constraints by rows
  f.con.K <- rbind(diag(12),diag(12))
  f.con.obs <- 
    matrix(c(rep(0,3),qCA/p00,0,qCC/p00,rep(0,3),qNC/p00,0,qNCp/p00,
             rep(0,7),qD/p01,0,qNC/p01,0,qNCp/p01,
             qA/p10,rep(0,5),qD/p10,rep(0,5),
             qA/p11,0,qCA/p11,0,qCC/p11,rep(0,7)), 
           ncol = 12, byrow = TRUE)
  f.con <- rbind(f.con.K, f.con.obs)
  
  # Set inequality signs
  f.dir.K <- c(rep("<=",12),rep(">=",12))
  f.dir.obs <- rep("==",4)
  f.dir <- c(f.dir.K, f.dir.obs)
  
  # Set right hand side coefficients
  f.rhs.K <- c(rep(K[2],12),rep(K[1],12))
  f.rhs.obs <- c(y00,y01,y10,y11)
  f.rhs <- c(f.rhs.K, f.rhs.obs)
  
  # Take care of Negative TE
  if (NegTE == TRUE) {
    f.con.NegTE <- 
      matrix(c(1,-1,rep(0,10),
               rep(0,2),1,-1,rep(0,8),
               rep(0,4),1,-1,rep(0,6),
               rep(0,6),1,-1,rep(0,4),
               rep(0,8),1,-1,rep(0,2),
               rep(0,10),1,-1), 
             ncol = 12, byrow = TRUE)
    f.dir.NegTE <- rep("<=",6)
    f.rhs.NegTE <- rep(0,6)
    
    #Replace constraint set
    f.con <- rbind(f.con,f.con.NegTE)
    f.dir <- c(f.dir,f.dir.NegTE)
    f.rhs <- c(f.rhs,f.rhs.NegTE)
  }
  
  if (is.na(minTE) == FALSE) {
    f.con.minTE <- 
      matrix(c(1,-1,rep(0,10),
               rep(0,2),1,-1,rep(0,8),
               rep(0,4),1,-1,rep(0,6),
               rep(0,6),1,-1,rep(0,4),
               rep(0,8),1,-1,rep(0,2),
               rep(0,10),1,-1), 
             ncol = 12, byrow = TRUE)
    f.dir.minTE <- rep(">=",6)
    f.rhs.minTE <- rep(minTE,6)
    
    #Replace constraint set
    f.con <- rbind(f.con,f.con.minTE)
    f.dir <- c(f.dir,f.dir.minTE)
    f.rhs <- c(f.rhs,f.rhs.minTE)
  }
  
  
  # LP
  lpsol.min <- lp("min", f.obj, f.con, f.dir, f.rhs)
  lpsol.max <- lp("max", f.obj, f.con, f.dir, f.rhs)
  
  if (sum(lpsol.min$status) == 2) {
    stop("Problem Infeasible.")
  } else {
    minval <- lpsol.min$objval; minsol <- lpsol.min$solution
    maxval <- lpsol.max$objval; maxsol <- lpsol.max$solution
    return(list(objval=c(minval,maxval), minsol=minsol, maxsol=maxsol))
  }
}

##############
# Include Covariates
# Input coefficient estimates from RF and FS
# xi are RF coefficients while gamma are FS coefficients
# WCoef is an indicator for whether to use coefficients on covariates
LATEstBoundsCov <- function(df,PrG6,beta,gam0,alphNCp,K=c(0,1),
                            NegTE=TRUE, minTE=-.3) {
  pobs <- empp(df)
  p00 <- pobs[1]; p10 <- pobs[2]; p01 <- pobs[3]; p11 <- pobs[4]
  qA <- PrG6[1]; qCA <- PrG6[2]; qCC <- PrG6[3]; qD <- PrG6[4]; 
  qNC <- PrG6[5]; qNCp <- PrG6[6]; 
  empz1 <- mean(df$z);
  
  # Force max and min K
  if (K[2]==Inf) K <- c(min(df$y),max(df$y))
  
  ## Build LP
  # 12-dim parameter to optimize over: etaA1, etaA0, etaCA1, etaCA0,
  # etaCC1, etaCC0, etaD1, etaD0, etaNC1, etaNC0, etaNN1, etaNN0
  
  # Choose Objective
  f.obj <- c(rep(0,4),qCC,-qCC,rep(0,2),qNC,-qNC,rep(0,2))/(qCC+qNC)
  
  
  # Set matrix corresponding to coefficients of constraints by rows
  f.con.K <- rbind(diag(12),diag(12))
  f.con.obs <- 
    matrix(c(0,0,qCA,-qCA,qCC,-qCC,-qD,qD,rep(0,4),
             0,0,0,qCA,0,qCC,0,0,0,qNC,0,alphNCp), 
           ncol = 12, byrow = TRUE)
  f.con <- rbind(f.con.K, f.con.obs)
  
  # Set inequality signs
  f.dir.K <- c(rep("<=",12),rep(">=",12))
  f.dir.obs <- rep("==",2)
  f.dir <- c(f.dir.K, f.dir.obs)
  
  # Set right hand side coefficients
  f.rhs.K <- c(rep(K[2],12),rep(K[1],12))
  f.rhs.obs <- c(beta*(qCA+qCC-qD),gam0)
  f.rhs <- c(f.rhs.K, f.rhs.obs)
  
  # Take care of Negative TE
  if (NegTE == TRUE) {
    f.con.NegTE <- 
      matrix(c(1,-1,rep(0,10),
               rep(0,2),1,-1,rep(0,8),
               rep(0,4),1,-1,rep(0,6),
               rep(0,6),1,-1,rep(0,4),
               rep(0,8),1,-1,rep(0,2),
               rep(0,10),1,-1), 
             ncol = 12, byrow = TRUE)
    f.dir.NegTE <- rep("<=",6)
    f.rhs.NegTE <- rep(0,6)
    
    #Replace constraint set
    f.con <- rbind(f.con,f.con.NegTE)
    f.dir <- c(f.dir,f.dir.NegTE)
    f.rhs <- c(f.rhs,f.rhs.NegTE)
  }
  
  if (is.na(minTE) == FALSE) {
    f.con.minTE <- 
      matrix(c(1,-1,rep(0,10),
               rep(0,2),1,-1,rep(0,8),
               rep(0,4),1,-1,rep(0,6),
               rep(0,6),1,-1,rep(0,4),
               rep(0,8),1,-1,rep(0,2),
               rep(0,10),1,-1), 
             ncol = 12, byrow = TRUE)
    f.dir.minTE <- rep(">=",6)
    f.rhs.minTE <- rep(minTE,6)
    
    #Replace constraint set
    f.con <- rbind(f.con,f.con.minTE)
    f.dir <- c(f.dir,f.dir.minTE)
    f.rhs <- c(f.rhs,f.rhs.minTE)
  }
  
  
  # LP
  lpsol.min <- lp("min", f.obj, f.con, f.dir, f.rhs)
  lpsol.max <- lp("max", f.obj, f.con, f.dir, f.rhs)
  
  if (sum(lpsol.min$status) == 2) {
    stop("Problem Infeasible.")
  } else {
    minval <- lpsol.min$objval; minsol <- lpsol.min$solution
    maxval <- lpsol.max$objval; maxsol <- lpsol.max$solution
    return(list(objval=c(minval,maxval), minsol=minsol, maxsol=maxsol))
  }
}
